//  mbed GBA Loader - Boot up a GameBoy Advance using Multiboot (no cart req.)
//  
//  Ken Kaarvik    kkaarvik@yahoo.com    Nov 13, 2010
//  akkera102                            Nov 08, 2014(modified)
//
//  mbed(LPC1768)   gba serial port(color)
//  p1-0V           6-GND (blue)
//  p5-mosi         3-SI  (orange)
//  p6-miso         2-SO  (red)
//  p7-sck          5-SC  (green)
//
//  The GBA file that you wish to load into the GBA is stored on the mbed's local file system.
//  Apply power to the mbed (or reset) before turning on the GBA.

#include "mbed.h"

LocalFileSystem local("local");
Serial pc(USBTX, USBRX);
SPI spi(p5, p6, p7);

DigitalOut myled1(LED1); // used for my demo program (volt4.gba)
DigitalOut myled2(LED2);
DigitalOut myled3(LED3);
DigitalOut myled4(LED4);


uint32_t writeSPI32NoDebug(uint32_t w)
{
    uint16_t w1 = w >> 16;
    uint16_t w2 = w & 0xffff;

    uint32_t r;
    r = spi.write(w1) << 16;
    r = spi.write(w2) | r;
    
    return r;
}

uint32_t writeSPI32(uint32_t w, char* msg)
{
    uint32_t r = writeSPI32NoDebug(w);

    pc.printf("0x%08x 0x%08x  ; %s\n",r , w, msg);
    return  r;
}


int main(void) 
{
    spi.format(16, 3);
    spi.frequency(100000);


    FILE *fp = fopen("/local/test.gba", "rb");  // change .gba file here
    if(fp == NULL)
    {
        pc.printf("Err: Can't open file\n");
        return EXIT_FAILURE;
    }

    fseek(fp, 0L, SEEK_END);
    long fsize = (ftell(fp) + 0x0f) & 0xfffffff0;

    if(fsize > 0x40000)
    {
        pc.printf("Err: Max file size 256kB\n");
        return EXIT_FAILURE;
    }

    fseek(fp, 0L, SEEK_SET);
    long fcnt = 0;


    uint32_t r, w, w2;
    int i, bit;

    do
    {
        r = writeSPI32(0x00006202, "Looking for GBA");
        wait(0.01);

    } while(r != 0x72026202);

    r = writeSPI32(0x00006202, "Found GBA");
    r = writeSPI32(0x00006102, "Recognition OK");

    pc.printf("Send Header(NoDebug)\n");
    for(i=0; i<=0x5f; i++)
    {
        w = getc(fp);
        w = getc(fp) << 8 | w;
        fcnt += 2;

        r = writeSPI32NoDebug(w);
    }

    r = writeSPI32(0x00006200, "Transfer of header data complete");
    r = writeSPI32(0x00006202, "Exchange master/slave info again");

    r = writeSPI32(0x000063d1, "Send palette data");
    r = writeSPI32(0x000063d1, "Send palette data, receive 0x73hh****");  

    uint32_t m = ((r & 0x00ff0000) >>  8) + 0xffff00d1;
    uint32_t h = ((r & 0x00ff0000) >> 16) + 0xf;

    r = writeSPI32((((r >> 16) + 0xf) & 0xff) | 0x00006400, "Send handshake data");
    r = writeSPI32((fsize - 0x190) / 4, "Send length info, receive seed 0x**cc****");

    uint32_t f = (((r & 0x00ff0000) >> 8) + h) | 0xffff0000;
    uint32_t c = 0x0000c387;


    pc.printf("Send encrypted data(NoDebug)\n");

    while(fcnt < fsize)
    {
        w = getc(fp);
        w = getc(fp) <<  8 | w;
        w = getc(fp) << 16 | w;
        w = getc(fp) << 24 | w;

        w2 = w;

        for(bit=0; bit<32; bit++)
        {
            if((c ^ w) & 0x01)
            {
                c = (c >> 1) ^ 0x0000c37b;
            }
            else
            {
                c = c >> 1;
            }

            w = w >> 1;
        }

        m = (0x6f646573 * m) + 1;
        writeSPI32NoDebug(w2 ^ ((~(0x02000000 + fcnt)) + 1) ^m ^0x43202f2f);

        fcnt = fcnt + 4;
    }
    fclose(fp);


    for(bit=0; bit<32; bit++)
    {
        if((c ^ f) & 0x01)
        {
            c =( c >> 1) ^ 0x0000c37b;
        }
        else
        {
            c = c >> 1;
        }

        f = f >> 1;
    }

    do
    {
        r = writeSPI32(0x00000065, "Wait for GBA to respond with CRC");

    } while(r != 0x00750065);

    r = writeSPI32(0x00000066, "GBA ready with CRC");
    r = writeSPI32(c,          "Let's exchange CRC!");

    pc.printf("CRC ...hope they match!\n");
    pc.printf("All done\n");


    for(;;)
    {
        r = writeSPI32NoDebug(0);
        wait(0.01);

        r =(~r) & 0x3ff;
//        myled1 = (r & 0x1); // turn on led1 when "A" button on GBA is pressed
//        myled2 = (r & 0x2); // B
//        myled3 = (r & 0x4); // Select
//        myled4 = (r & 0x8); // Start
    }
}
